package org.bouncycastle.jcajce.drbgprov;

import java.security.NoSuchAlgorithmException;
import java.security.Provider;
import java.security.SecureRandom;
import java.security.SecureRandomSpi;

import org.bouncycastle.crypto.CryptoServicesRegistrar;
import org.bouncycastle.crypto.EntropySource;
import org.bouncycastle.crypto.EntropySourceProvider;
import org.bouncycastle.crypto.fips.FipsDRBG;
import org.bouncycastle.util.Arrays;
import org.bouncycastle.util.Pack;
import org.bouncycastle.util.Strings;

/**
 * This class builds a single DRBG using the FIPS module and registers it.
 * <p>
 * It's largely useful when you're dealing with code that has a lot of "new SecureRandom()" calls
 * in it as "new SecureRandom()" results in the JVM index all the providers and looking for the first
 * SecureRandom definition it finds. If you put this provider first, what is defined in here is what
 * the JVM will use for "new SecureRandom()".
 * </p>
 * <p>
 * As SecureRandom is not part of the JCE this provider does not need to be signed using a JCE signing
 * certificate, so you can safely adjust as required.
 * </p>
 */
public class BouncyCastleDefaultDrbgFipsProvider
    extends Provider
{
    public BouncyCastleDefaultDrbgFipsProvider()
    {
        super("BCFIPS-DRBG", 1.0001, "Bouncy Castle Default DRBG Provider (1.0.1)");

        EntropySourceProvider entropyProvider;
        byte[] nonce;

        // the first section builds the entropy provider for the DRBG we are
        // creating - Java 8 and later will use SecureRandom.getInstanceStrong() and
        // Java 7 and earlier will use an internal Sun class (not ideal, but it's the only
        // way to get easy access to the seed generation - generateSeed() on both
        // the alternatives should take data from /dev/random, or its nearest equivalent,
        // which is a prediction resistant source.
        try
        {
            final SecureRandom random = SecureRandom.getInstanceStrong();

            // we generate a 16 byte nonce to add further differentiation to this DRBG
            // from any we have generated previously.
            nonce = random.generateSeed(16);

            entropyProvider = new EntropySourceProvider()
            {
                public EntropySource get(final int bitsRequired)
                {
                    return new EntropySource()
                    {
                        public boolean isPredictionResistant()
                        {
                            return true;
                        }

                        public byte[] getEntropy()
                        {
                            return random.generateSeed((bitsRequired + 7) / 8);
                        }

                        public int entropySize()
                        {
                            return bitsRequired;
                        }
                    };

                }
            };
        }
        catch (NoSuchAlgorithmException e)
        {
            final sun.security.provider.SecureRandom random = new sun.security.provider.SecureRandom();

            // we generate a 16 byte nonce to add further differentiation to this DRBG
            // from any we have generated previously.
            nonce = random.engineGenerateSeed(16);

            entropyProvider = new EntropySourceProvider()
            {
                public EntropySource get(final int bitsRequired)
                {
                    return new EntropySource()
                    {
                        public boolean isPredictionResistant()
                        {
                            return true;
                        }

                        public byte[] getEntropy()
                        {
                            return random.engineGenerateSeed((bitsRequired + 7) / 8);
                        }

                        public int entropySize()
                        {
                            return bitsRequired;
                        }
                    };
                }
            };
        }

        // set up additional differentiation using a personalisation string.
        byte[] personalizationString = Arrays.concatenate(Strings.toByteArray("Define DRBG Default"),
                    Pack.longToLittleEndian(Thread.currentThread().getId()), Pack.longToLittleEndian(System.currentTimeMillis()));

        // we now use the nonce, personalizationString, and the entropy provider
        // we have created to build a DRBG based on SHA-512. This DRBG does not
        // have prediction resistance set so it will stretch the seed data from the
        // entropy provider until it needs to reseed.
        final SecureRandom random = FipsDRBG.SHA512
            .fromEntropySource(entropyProvider)
            .setPersonalizationString(personalizationString)
            .build(nonce,
                false, Strings.toByteArray("Bouncy Castle DRBG without PR"));

        // install this as our default random as well.
        CryptoServicesRegistrar.setSecureRandom(random);

        // install in the provider so the JCA can find it.
        put("SecureRandom.GENERAL", BouncyCastleDefaultDrbgFipsProvider.class.getName() + "$Default");
    }

    /**
     * Default is just a wrapper around the default random. This will get lazilly
     * built when the JVM needs it, preventing a "new SecureRandom()" from causing
     * a panic.
     */
    public static class Default
        extends SecureRandomSpi
    {
        private static final SecureRandom random = CryptoServicesRegistrar.getSecureRandom();

        public Default()
        {
        }

        protected void engineSetSeed(byte[] bytes)
        {
            random.setSeed(bytes);
        }

        protected void engineNextBytes(byte[] bytes)
        {
            random.nextBytes(bytes);
        }

        protected byte[] engineGenerateSeed(int numBytes)
        {
            return random.generateSeed(numBytes);
        }
    }
}
